package gnu.brl;

// brlsv.java -- the BRL servlet
// Copyright (C) 1999  Bruce R. Lewis and Eaton Vance Management
// See the file COPYING for license terms.

import java.io.*;
import java.util.*;

import javax.servlet.*;
import javax.servlet.http.*;

import kawa.lang.*;
import kawa.standard.*;
import gnu.mapping.*;
import gnu.expr.*;
import gnu.kawa.util.*;

public class brlsv extends HttpServlet
{
    // One interpreter shared by all instances
    static BRL interp = new BRL();

    // Constants for state
    static final byte STATE_UNINITIALIZED=0;
    static final byte STATE_NORMAL=1;	// handling requests normally
    static final byte STATE_GLOBALERROR=2; // error in global .scm file
    static final byte STATE_NOSCMDIR=3; // missing scmdir init parameter

    static byte interp_status = STATE_UNINITIALIZED;

    static String problem_fname; // filename of global .scm with error
    static Throwable problem_throw;
    static String global_fname;	// brl-global.scm
    static long global_modified;
    static String site_fname;	// sitedefs.scm
    static long site_modified;

    void init()
	throws java.io.IOException
    {
	synchronized (interp) {
	    String scmdir = getInitParameter("scmdir");
	    if (scmdir == null)
		interp_status = STATE_NOSCMDIR;
	    else {
		interp_status = STATE_NORMAL;
		scmdir = scmdir + File.separator;
		interp.define("brl-scmdir", new FString(scmdir));
		global_fname = scmdir + "brl-global.scm";
		site_fname = scmdir + "sitedefs.scm";
		global_modified =
		    reloadScmFile(global_fname, global_modified);
		site_modified =
		    reloadScmFile(site_fname, site_modified);
	    }
	}
    }

    long reloadScmFile(String fname, long lastmod) {

	long new_modified = (new File(fname)).lastModified();
	if (new_modified != lastmod) {
	    try {
		interp.eval((Object)
			    new Pair("load", new Pair(new FString(fname),
						      LList.Empty)),
			    interp.getEnvironment());
		if (problem_fname == fname)
		    interp_status = STATE_NORMAL; // problem fixed
	    }
	    catch (Throwable ex) {
		interp_status = STATE_GLOBALERROR;
		problem_fname = fname;
		problem_throw = ex;
	    }
	}
	return new_modified;
    }

    public void doPost (HttpServletRequest req, HttpServletResponse res)
	throws ServletException, IOException
    {
	doGet(req, res);
    }

    public void doGet (HttpServletRequest req, HttpServletResponse res)
	throws ServletException, IOException
    {
	if (global_fname != null && site_fname != null) {
	    // Reload global and site .scm files if necessary
	    synchronized (global_fname) {
		global_modified = reloadScmFile(global_fname, global_modified);
	    }
	    if (interp_status != STATE_GLOBALERROR
		&& !(global_fname.equals(problem_fname)))
		synchronized (site_fname) {
		    site_modified = reloadScmFile(site_fname, site_modified);
		}
	} else {
	    init();		// Needs explicit call with JSDK2.0
	}

	// Handle request according to state
	switch(interp_status) {
	case STATE_GLOBALERROR:
	    doGlobalError(req, res);
	    return;
	case STATE_NOSCMDIR:
	    doNoScmError(req, res);
	    return;
	default:
	    doNormal(req, res);
	    return;
	}
    }

    public void doGlobalError (HttpServletRequest req, HttpServletResponse res)
	throws ServletException, IOException
    {
	res.setContentType("text/plain");
	PrintWriter toclient = res.getWriter();

	toclient.println("Request failed due to an error in a global file:");
	toclient.println(problem_fname);
	toclient.println(problem_throw.getMessage());
	problem_throw.printStackTrace(toclient);
	return;
    }

    public void doNoScmError (HttpServletRequest req, HttpServletResponse res)
	throws ServletException, IOException
    {
	res.setContentType("text/plain");
	PrintWriter toclient = res.getWriter();

	toclient.println("This servlet initialization parameter must be set:");
	toclient.println("\tscmdir");
	return;
    }

    public void doNormal (HttpServletRequest req, HttpServletResponse res)
	throws ServletException, IOException
    {
	PrintWriter toclient;
	Object schemeobj;
	boolean response_started=false;
	String requested_fname;

	if (req.getPathInfo() == null)
	    requested_fname = req.getRealPath(req.getRequestURI());
	else
	    requested_fname = req.getPathTranslated();

	// Create s-expr to handle request
	Object s_expr =
	    LList.makeList(new Object[] { "brl-handle-request",
					 new FString(requested_fname),
					 LList.list2("quote",
						    getSchemeBindings(req)),
					 LList.list2("quote", "servlet"),
					 req,
					 res }, 0);

	// Send output
	try
	    {
		schemeobj=interp.eval((Object) s_expr, interp.getEnvironment());
		response_started=true;
		res.setContentType("text/html");
		toclient = res.getWriter();
		toclient.print(schemeobj.toString());
	    }
	catch (Throwable ex)
	    {
		if (!response_started)
		    {
			res.setContentType("text/plain");
			toclient = res.getWriter();
			toclient.println("Error message:");
			toclient.println(ex.getMessage());
			toclient.println("Stack trace:");
			ex.printStackTrace(toclient);
		    }
	    }
    }

    // Return either a Scheme string or list of Scheme strings
    public static Object getSchemeValues(String[] jstrings)
    {
	if (jstrings.length == 1)
	    return new FString(jstrings[0]);

	LList result = LList.Empty;
	for (int i = jstrings.length;  --i >= 0; )
	    result = new Pair (new FString(jstrings[i]), result);
	return result;
    }

    // Return list of bindings (Scheme symbol/string pairs)
    public static Object getSchemeBindings(HttpServletRequest req)
    {
	Enumeration e;
	LList bindings = LList.Empty;
	String name;
	Object values;

	e = req.getParameterNames();
	while (e.hasMoreElements())
	    {
		name = ((String)e.nextElement()).intern();
		values = getSchemeValues(req.getParameterValues(name));
		bindings = new Pair(new Pair(name, values), bindings);
	    }
	return bindings;
    }

    // Return a Scheme list of cookies
    public static Object getSchemeCookies(HttpServletRequest req)
    {
	Cookie[] carray = req.getCookies();
	if (carray == null || carray.length==0) return LList.Empty;

	Pair clist = (Pair)LList.makeList((Object[]) req.getCookies(), 0);
	Pair p;
	Cookie c;

	for (Object o = clist; o != LList.Empty; o = ((Pair)o).cdr)
	    {
		p = (Pair) o;
		c = (Cookie) p.car;
		p.car = LList.list3(c.getName(),
				   new FString(c.getValue()),
				   c);
	    }
	return clist;
    }

    public String getServletInfo()
    {
	return
	    Version.package_string + " servlet " + Version.release_string;
    }
}
